#define G_LOG_DOMAIN "boilerplate-plugins-engine"

#include "config.h"

#include "boilerplate-plugins-engine.h"
#include "boilerplate-activatable.h"
#include "boilerplate-application.h"

#include <glib.h>
#include <libpeas/peas.h>

typedef struct _BoilerplatePluginsEnginePrivate
{
    GSettings *settings;

    GHashTable *extensions;

} BoilerplatePluginsEnginePrivate;

struct _BoilerplatePluginsEngine
{
    PeasEngine parent_instance;
};

/**
 * SECTION:boilerplate-plugins-engine
 * @title: BoilerplatePluginsEngine
 * @short_description: Wrapper object for initializing plugins
 * @include: boilerplate-plugins-engine.h
 * @see_also: #PeasEngine
 *
 * #BoilerplatePluginsEngine is a wrapper object for initializing the #PeasEngine.
 **/

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplatePluginsEngine, boilerplate_plugins_engine, PEAS_TYPE_ENGINE)

enum
{
  PLUGIN_LOADED,
  PLUGIN_UNLOADED,
  NUM_SIGNALS
};

static guint signals[NUM_SIGNALS] = { 0, };

/* Singleton instance */
BoilerplatePluginsEngine *default_engine = NULL;

static void
on_plugin_loaded (BoilerplatePluginsEngine *self,
                  PeasPluginInfo           *info)
{
    // TODO: Plugins Debug - Remove when done.
    g_print ("Plugins Engine: extension added\n");

    if (peas_engine_provides_extension (PEAS_ENGINE (self), info, BOILERPLATE_TYPE_ACTIVATABLE))
    {
        BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
        BoilerplateActivatable *activatable;
        PeasExtension *extension;

        extension = peas_engine_create_extension (PEAS_ENGINE (self),
                                                  info,
                                                  BOILERPLATE_TYPE_ACTIVATABLE,
                                                  NULL);

        g_hash_table_insert (priv->extensions,
                             info,
                             extension);

        activatable = BOILERPLATE_ACTIVATABLE (extension);

        boilerplate_activatable_activate (activatable);

        g_signal_emit (self, signals[PLUGIN_LOADED], 0, info, extension);
    }
}

static void
on_plugin_unloaded (BoilerplatePluginsEngine *self,
                    PeasPluginInfo           *info)
{
    // TODO: Plugins Debug - Remove when done.
    g_print ("Plugins Engine: extension removed\n");

    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
    BoilerplateActivatable *activatable;

    activatable = g_hash_table_lookup (priv->extensions, info);

    if (!activatable)
        return;

    boilerplate_activatable_deactivate (activatable);

    g_signal_emit (self, signals[PLUGIN_UNLOADED], 0, info, activatable);

    g_hash_table_remove (priv->extensions, info);

    g_clear_object (&activatable);
}

static void
boilerplate_plugins_engine_dispose (GObject *object)
{
    BoilerplatePluginsEngine *self = BOILERPLATE_PLUGINS_ENGINE (object);
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    if (priv->settings != NULL)
    {
        g_object_unref (priv->settings);
    }
    priv->settings = NULL;

    G_OBJECT_CLASS (boilerplate_plugins_engine_parent_class)->dispose (object);
}

static void
boilerplate_plugins_engine_finalize (GObject *object)
{
    BoilerplatePluginsEngine *self = BOILERPLATE_PLUGINS_ENGINE (object);
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    g_clear_pointer (&priv->extensions, g_hash_table_destroy);

    G_OBJECT_CLASS (boilerplate_plugins_engine_parent_class)->finalize (object);
}
static void
boilerplate_plugins_engine_init (BoilerplatePluginsEngine *self)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    /* Setup basic variables */
    priv->settings = g_settings_new (PACKAGE_ID ".plugins");
    priv->extensions = g_hash_table_new (g_direct_hash, g_direct_equal);

    // TODO: Plugins Debug - Remove when done.
    g_print ("Plugins Engine: Plugins init\n");

    peas_engine_enable_loader (PEAS_ENGINE (self), "python3");
    peas_engine_enable_loader (PEAS_ENGINE (self), "lua5.1");

    /* Add default plugins dir */
    peas_engine_add_search_path (PEAS_ENGINE (self),
	                             PACKAGE_PLUGINS_DIRECTORY,
                                 PACKAGE_PLUGINS_DATA_DIRECTORY);

    peas_engine_add_search_path (PEAS_ENGINE (self),
	                             g_build_filename (g_get_home_dir (),
                                                   ".local",
                                                   "share",
                                                   PACKAGE,
                                                   "plugins",
                                                   NULL),
	                             NULL);

    // TODO: Plugins Debug - Remove when done.
    g_print ("Plugins Engine: Added plugin search path: %s\n", PACKAGE_PLUGINS_DIRECTORY);
    g_print ("Plugins Engine: Added plugin data search path: %s\n", PACKAGE_PLUGINS_DATA_DIRECTORY);
    g_print ("Plugins Engine: Added plugin search path: %s\n", g_build_filename (g_get_home_dir (),
                                                                                 ".local",
                                                                                 "share",
                                                                                 PACKAGE,
                                                                                 "plugins",
                                                                                 NULL));

    /* (un)load-ers */
    g_signal_connect_after (self,
                            "load-plugin",
                            G_CALLBACK (on_plugin_loaded),
                            NULL);

    g_signal_connect (self,
                      "unload-plugin",
                      G_CALLBACK (on_plugin_unloaded),
                      NULL);

    /* Load saved plugins */
    g_settings_bind (priv->settings,
                     "active-extensions",
                     self,
                     "loaded-plugins",
                     G_SETTINGS_BIND_DEFAULT);

    const gchar *load[] = {"boilerplate-dark-theme", NULL};

    peas_engine_set_loaded_plugins (PEAS_ENGINE (self), load);
}

static void
boilerplate_plugins_engine_class_init (BoilerplatePluginsEngineClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = boilerplate_plugins_engine_dispose;
    object_class->finalize = boilerplate_plugins_engine_finalize;


    signals[PLUGIN_LOADED] = g_signal_new ("plugin-loaded",
                                           BOILERPLATE_TYPE_PLUGINS_ENGINE,
                                           G_SIGNAL_RUN_FIRST,
                                           0,
                                           NULL,
                                           NULL,
                                           NULL,
                                           G_TYPE_NONE,
                                           2,
                                           PEAS_TYPE_PLUGIN_INFO,
                                           BOILERPLATE_TYPE_ACTIVATABLE);

    signals[PLUGIN_UNLOADED] = g_signal_new ("plugin-unloaded",
                                             BOILERPLATE_TYPE_PLUGINS_ENGINE,
                                             G_SIGNAL_RUN_FIRST,
                                             0,
                                             NULL,
                                             NULL,
                                             NULL,
                                             G_TYPE_NONE,
                                             2,
                                             PEAS_TYPE_PLUGIN_INFO,
                                             BOILERPLATE_TYPE_ACTIVATABLE);
}

/**
 * boilerplate_plugins_engine_new:
 *
 * Creates a new #PeasEngine instance as #BoilerplatePlguinsEngine.
 *
 * Returns: #BoilerplatePlguinsEngine
 */
BoilerplatePluginsEngine *
boilerplate_plugins_engine_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_PLUGINS_ENGINE,
                         NULL);
}

/**
 * boilerplate_plugins_engine_get_default:
 *
 * Gets the default #PeasEngine, if does not exist create a new instance of
 * #BoilerplatePlguinsEngine.
 *
 * Returns: #BoilerplatePlguinsEngine
 */
BoilerplatePluginsEngine *
boilerplate_plugins_engine_get_default (void)
{
    if (!default_engine)
    {
        default_engine = boilerplate_plugins_engine_new ();
    }

    return default_engine;
}

