#define G_LOG_DOMAIN "boilerplate-preferences-view"

#include "config.h"

#include "boilerplate-preferences-view.h"

#include <gtk/gtk.h>

/**
 * SECTION:boilerplate-preferences-view
 * @title: BoilerplatePreferencesView
 * @short_description: A #GtkBin instance created for #BoilerplatePreferences
 * @include: boilerplate-preferences-view.h
 * @see_also: #GtkBin, #BoilerplatePreferencesView
 *
 * The #BoilerplatePreferencesView is an instance #GtkBin, it's the base object
 * for the preferences options. It contains a #GtkStack and manages the adding,
 * removing, and update of options in the preferences view. This allows for the
 * back-end code for managing the preferences ui to be added to either a
 * #GtkWindow, #GtkDialog or other custom #GtkWidget without re-writing an
 * entire GObject.
 **/

typedef struct _BoilerplatePreferencesViewPrivate
{
    GHashTable *pages;
    GHashTable *groups;
} BoilerplatePreferencesViewPrivate;

struct _BoilerplatePreferencesView
{
    GtkBin parent_instance;

    GtkWidget *view;
    GtkWidget *sidebar;
    GtkWidget *pages;
};

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplatePreferencesView, boilerplate_preferences_view, GTK_TYPE_BIN)

/**
 * boilerplate_preferences_view_add_page:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 * @page_title: (gchar *): The name of the page that will be shown in the
 * #GtkStackSidebar
 *
 * Adds a page with @page_name to the current @self instance.
 */
void
boilerplate_preferences_view_add_page (BoilerplatePreferencesView *self,
                                       const gchar                *page_name,
                                       const gchar                *page_title)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);
    GtkWidget *page;

    g_return_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self));
    g_return_if_fail (page_name != NULL);
    g_return_if_fail (page_title != NULL);

    page = g_object_new (GTK_TYPE_BOX,
                         "visible", TRUE,
                         "expand", TRUE,
                         NULL);

    if (page == NULL)
    {
        g_warning ("New page could not be created.");
        return;
    }

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return;
    }

    if (page_title == NULL)
    {
        g_warning ("Page title can not be NULL.");
        return;
    }

    gtk_stack_add_titled (GTK_STACK (self->pages),
                          GTK_WIDGET (page),
                          page_name,
                          page_title);

    g_hash_table_insert (priv->pages, g_strdup (page_name), page);
}

/**
 * boilerplate_preferences_view_get_page:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 *
 * Gets a page with @page_name from the #GtkStack in @self instance.
 *
 * Returns: (GtkWidget *) (nullable): #GtkWidget of @page_name
 */
GtkWidget *
boilerplate_preferences_view_get_page (BoilerplatePreferencesView *self,
                                       const gchar                *page_name)
{
    return gtk_stack_get_child_by_name (GTK_STACK (self->pages), page_name);
}

/**
 * boilerplate_preferences_view_add_group:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 * @group_name: (gchar *): The internal name of the group, similar to an object
 * ID. No special characters, spaces, etc..
 * @group_title: (gchar *): The name of the group that will be shown above the
 * group's #GtkFrame and #GtkListBox
 *
 * Adds a group with @group_name to the @page_name page in the current @self
 * instance.
 */
void
boilerplate_preferences_view_add_group (BoilerplatePreferencesView *self,
                                       const gchar                 *page_name,
                                       const gchar                 *group_name,
                                       const gchar                 *group_title)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    BoilerplatePreferencesGroup *group;

    GtkWidget *page,
              *contents,
              *title,
              *frame,
              *scrolled_window,
              *list_box;

    PangoAttrList *attributes;
    PangoAttribute *attribute;

    group = g_new (BoilerplatePreferencesGroup, 1);

    page = boilerplate_preferences_view_get_page (self, page_name);

    if (page == NULL)
    {
        g_warning ("Page with name \"%s\" could not be found.", page_name);
        return;
    }

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return;
    }

    if (group_title == NULL)
    {
        g_warning ("Group title can not be NULL.");
        return;
    }

    contents = g_object_new (GTK_TYPE_BOX,
                          "visible", TRUE,
                          "orientation", GTK_ORIENTATION_VERTICAL,
                          "spacing", 8,
                          NULL);

    attributes = pango_attr_list_new ();
    attribute = pango_attr_weight_new (PANGO_WEIGHT_HEAVY);
    pango_attr_list_insert (attributes, attribute);

    title = g_object_new (GTK_TYPE_LABEL,
                          "label", group_title,
                          "visible", TRUE,
                          "hexpand", TRUE,
                          "xalign", 0.0,
                          "attributes", attributes,
                          NULL);

    gtk_container_add (GTK_CONTAINER (contents),
                       GTK_WIDGET (title));

    frame = g_object_new (GTK_TYPE_FRAME,
                          "visible", TRUE,
                          "hexpand", TRUE,
                          NULL);

    gtk_container_add (GTK_CONTAINER (contents),
                       GTK_WIDGET (frame));

    scrolled_window = g_object_new (GTK_TYPE_SCROLLED_WINDOW,
                                    "visible", TRUE,
                                    "shadow-type", GTK_SHADOW_NONE,
                                    "propagate-natural-height", TRUE,
                                    NULL);

    gtk_container_add (GTK_CONTAINER (frame),
                       GTK_WIDGET (scrolled_window));

    list_box = g_object_new (GTK_TYPE_LIST_BOX,
                             "visible", TRUE,
                             "hexpand", TRUE,
                             "selection-mode", FALSE,
                             NULL);

    gtk_container_add (GTK_CONTAINER (scrolled_window),
                       GTK_WIDGET (list_box));

    gtk_container_add (GTK_CONTAINER (page),
                       GTK_WIDGET (contents));

    group->title = title;
    group->box = contents;
    group->list_box = list_box;

    g_hash_table_insert (priv->groups, g_strdup_printf ("%s.%s", page_name, group_name), group);
}

/**
 * boilerplate_preferences_view_get_group:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 *
 * Returns a #BoilerplatePreferencesGroup which contains a struct with references
 * to the #GtkLabel or #GtkListBox.
 *
 * Returns: (BoilerplatePreferencesGroup *) (nullable): #GtkWidget of @page_name
 */
BoilerplatePreferencesGroup *
boilerplate_preferences_view_get_group (BoilerplatePreferencesView *self,
                                        const gchar                *page_name,
                                        const gchar                *group_name)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);
    GtkWidget *group;

    g_return_val_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self), NULL);

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return NULL;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return NULL;
    }

    group = g_hash_table_lookup (priv->groups, g_strdup_printf ("%s.%s", page_name, group_name));

    if (group == NULL)
    {
        g_warning ("No group named \"%s\" in pages hash table.", group_name);
        return NULL;
    }

    return (BoilerplatePreferencesGroup *) group; // TODO: Avoid casting?
}

/**
 * boilerplate_preferences_view_add_row:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 * @group_name: (gchar *): The internal name of the group, similar to an object
 * ID. No special characters, spaces, etc..
 *
 * Adds a #GtkListBoxRow to @group_name in the @page_name.
 *
 * Returns (GtkWidget *) (nullable): References to the row created
 */
GtkWidget *
boilerplate_preferences_view_add_row (BoilerplatePreferencesView *self,
                                      const gchar                *page_name,
                                      const gchar                *group_name)
{
    BoilerplatePreferencesGroup *group;
    GtkWidget *row;

    g_return_val_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self), NULL);

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return NULL;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return NULL;
    }

    group = boilerplate_preferences_view_get_group (self, page_name, group_name);

    if (group == NULL)
    {
        g_warning ("Group can not be null.");
        return NULL;
    }

    row = g_object_new (GTK_TYPE_LIST_BOX_ROW,
                        "visible", TRUE,
                        "hexpand", TRUE,
                        NULL);

    gtk_container_add (GTK_CONTAINER (group->list_box),
                       GTK_WIDGET (row));

    return row;
}

/**
 * boilerplate_preferences_view_add_option:
 * @self: A instance of #BoilerplatePreferencesView
 * @page_name: (gchar *): The internal name of the page, similar to an object
 * ID. No special characters, spaces, etc..
 * @group_name: (gchar *): The internal name of the group, similar to an object
 * ID. No special characters, spaces, etc..
 * @title: The name of the option
 * @subtitle: The description of the option
 * @widget: (GtkWidget *) (nullable): A #GtkWidget to add to the row.
 *
 * Creates a new @group_name row in @page_name with information like @title,
 * @subtitle, & @widget.
 */
void
boilerplate_preferences_view_add_option (BoilerplatePreferencesView *self,
                                         const gchar                *page_name,
                                         const gchar                *group_name,
                                         const gchar                *title,
                                         const gchar                *subtitle,
                                         GtkWidget                  *widget)
{
    GtkWidget *row,
              *contents,
              *info,
              *title_label,
              *subtitle_label;

    PangoAttrList *attributes;
    PangoAttribute *attribute;

    g_return_if_fail (BOILERPLATE_IS_PREFERENCES_VIEW (self));

    if (page_name == NULL)
    {
        g_warning ("Page name can not be NULL.");
        return;
    }

    if (group_name == NULL)
    {
        g_warning ("Group name can not be NULL.");
        return;
    }

    if (title == NULL)
    {
        g_warning ("Title can not be NULL.");
        return;
    }

    if (subtitle == NULL)
    {
        g_warning ("Subtitle can not be NULL.");
        return;
    }

    if (widget == NULL)
    {
        g_warning ("Widget can not be NULL and must be a GtkWidget.");
        return;
    }

    row = boilerplate_preferences_view_add_row (self, page_name, group_name);

    contents = g_object_new (GTK_TYPE_BOX,
                             "visible", TRUE,
                             "hexpand", TRUE,
                             "orientation", GTK_ORIENTATION_HORIZONTAL,
                             "margin", 10,
                             NULL);

    info  = g_object_new (GTK_TYPE_BOX,
                             "visible", TRUE,
                             "expand", TRUE,
                             "orientation", GTK_ORIENTATION_VERTICAL,
                             NULL);

    gtk_container_add (GTK_CONTAINER (contents),
                       GTK_WIDGET (info));

    title_label = g_object_new (GTK_TYPE_LABEL,
                                "label", title,
                                "visible", TRUE,
                                "hexpand", TRUE,
                                "halign", GTK_ALIGN_START,
                                NULL);

    gtk_container_add (GTK_CONTAINER (info),
                       GTK_WIDGET (title_label));

    attributes = pango_attr_list_new ();
    attribute = pango_attr_scale_new (0.83333);
    pango_attr_list_insert (attributes, attribute);

    subtitle_label = g_object_new (GTK_TYPE_LABEL,
                                   "label", subtitle,
                                   "visible", TRUE,
                                   "hexpand", TRUE,
                                   "halign", GTK_ALIGN_START,
                                   "wrap", TRUE,
                                   "attributes", attributes,
                                   NULL);

    gtk_container_add (GTK_CONTAINER (info),
                       GTK_WIDGET (subtitle_label));

    gtk_container_add (GTK_CONTAINER (contents),
                       GTK_WIDGET (widget));

    gtk_container_add (GTK_CONTAINER (row),
                       GTK_WIDGET (contents));
}

/*
 * boilerplate_preferences_view_finalize:
 * @object: An instance of #BoilerplatePreferencesView
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
boilerplate_preferences_view_finalize (GObject *object)
{
    BoilerplatePreferencesView *self = BOILERPLATE_PREFERENCES_VIEW (object);
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    g_clear_pointer (&priv->pages, g_hash_table_unref);
    g_clear_pointer (&priv->groups, g_hash_table_unref);

    G_OBJECT_CLASS (boilerplate_preferences_view_parent_class)->finalize (object);
}

/*
 * boilerplate_preferences_view_init:
 * @self: An instance of #BoilerplatePreferencesView
 *
 * Initializes various aspects of the window. Anything the window would need
 * before becoming presented should be done here.
 */
static void
boilerplate_preferences_view_init (BoilerplatePreferencesView *self)
{
    BoilerplatePreferencesViewPrivate* priv = boilerplate_preferences_view_get_instance_private (self);

    gtk_widget_init_template (GTK_WIDGET (self));

    priv->pages = g_hash_table_new (g_str_hash, g_str_equal);
    priv->groups = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);
}

/*
 * boilerplate_preferences_view_class_init:
 * @klass: An instance of #BoilerplatePreferencesViewClass
 *
 * Initializes the various virtural functions in #BoilerplatePreferencesViewClass.
 * In addtition it connects the various #GtkWidgets from the gresource file
 * using gtk_widget_class_set_template_from_resource().
 */
static void
boilerplate_preferences_view_class_init (BoilerplatePreferencesViewClass *klass)
{
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = boilerplate_preferences_view_finalize;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/preferences/view.ui");

    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, view);
    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, sidebar);
    gtk_widget_class_bind_template_child (widget_class, BoilerplatePreferencesView, pages);
}

/**
 * boilerplate_preferences_view_new:
 *
 * Creates a new #BoilerplatePreferencesView, a sub-object of #GtkBin.
 *
 * ## Usage
 * |[
 * BoilerplatePreferencesView *preferences_view;
 *
 * preferences_view = boilerplate_preferences_view_new ();
 * ]|
 *
 * Returns: #BoilerplatePreferencesView; a sub-object of #GtkBin
 */
BoilerplatePreferencesView *
boilerplate_preferences_view_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_PREFERENCES_VIEW,
                         NULL);
}

